// dma.h
#pragma once
#include <stdint.h>

typedef enum dma_burst_enum {
  BURST_SINGLE,
  BURST_INCR4,
  BURST_INCR8,
  BURST_INCR16
} dma_burst_enum;

typedef enum dma_priority_enum {
  PRIO_LOW,
  PRIO_MEDIUM,
  PRIO_HIGH,
  PRIO_VERY_HIGH
} dma_priority_enum;

typedef enum dma_data_size_enum {
  DSIZE_BYTE,
  DSIZE_HALF_WORD,
  DSIZE_WORD
} dma_data_size_enum;

typedef enum dma_transfer_direction_enum {
  DDIR_PERI_TO_MEM,
  DDIR_MEM_TO_PERI,
  DDIR_MEM_TO_MEM
} dma_transfer_direction_enum;

typedef volatile struct /* dma_stream_struct */ {  // word access
// 00 Stream x configuration register (DMA_SxCR)
  uint32_t  CR_EN     :  1; // Stream enable, flag stream ready when read low
  uint32_t  CR_DMEIE  :  1; // Direct mode error interrupt enable
  uint32_t  CR_TEIE   :  1; // Transfer error interrupt enable
  uint32_t  CR_HTIE   :  1; // Half transfer interrupt enable
  uint32_t  CR_TCIE   :  1; // Transfer complete interrupt enable
  uint32_t  CR_PFCTRL :  1; // Peripheral flow controller
  uint32_t  CR_DIR    :  2; // Data transfer direction
  uint32_t  CR_CIRC   :  1; // Circular mode
  uint32_t  CR_PINC   :  1; // Peripheral increment mode
  uint32_t  CR_MINC   :  1; // Memory increment mode
  uint32_t  CR_PSIZE  :  2; // Peripheral data size
  uint32_t  CR_MSIZE  :  2; // Memory data size
  uint32_t  CR_PINCOS :  1; // Peripheral increment offset size
  uint32_t  CR_PL     :  2; // Priority level
  uint32_t  CR_DBM    :  1; // Double buffer mode
  uint32_t  CR_CT     :  1; // Current target (only in double buffer mode)
  uint32_t            :  1;
  uint32_t  CR_PBURST :  2; // Peripheral burst transfer configuration
  uint32_t  CR_MBURST :  2; // Memory burst transfer configuration
  uint32_t  CR_CHSEL  :  3; // Channel selection
  uint32_t            :  4;
// 04 Stream x number of data register (DMA_SxNDTR)
  uint32_t  NDTR;           // 0 ... 65535
// 08 Stream x peripheral address register (DMA_SxPAR)
  uint32_t  PAR;
// 0C Stream x memory 0 address register (DMA_SxM0AR)
  uint32_t  M0AR;
// 10 Stream x memory 1 address register (DMA_SxM1AR)
  uint32_t  M1AR;
// 14 Stream x FIFO control register (DMA_SxFCR)
  uint32_t  FCR_FTH   :  2; // FIFO threshold selection
  uint32_t  FCR_DMDIS :  1; // Direct mode disable
  uint32_t  FCR_FS    :  3; // FIFO status
  uint32_t              :  1;
  uint32_t  FCR_FEIE  :  1; // FIFO error interrupt enable
  uint32_t              : 24;
} dma_stream_struct;


typedef volatile struct /* dma_struct */ {  // word access
// 00 Low interrupt status register (DMA_LISR)
  uint32_t  LISR_FEIF0    :  1; // Stream 0 FIFO error interrupt
  uint32_t                :  1;
  uint32_t  LISR_DMEIF0   :  1; // Stream 0 direct mode error interrupt
  uint32_t  LISR_TEIF0    :  1; // Stream 0 transfer error interrupt
  uint32_t  LISR_HTIF0    :  1; // Stream 0 half transfer interrupt
  uint32_t  LISR_TCIF0    :  1; // Stream 0 transfer complete interrupt
  uint32_t  LISR_FEIF1    :  1; // Stream 1 FIFO error interrupt
  uint32_t                :  1;
  uint32_t  LISR_DMEIF1   :  1; // Stream 1 direct mode error interrupt
  uint32_t  LISR_TEIF1    :  1; // Stream 1 transfer error interrupt
  uint32_t  LISR_HTIF1    :  1; // Stream 1 half transfer interrupt
  uint32_t  LISR_TCIF1    :  1; // Stream 1 transfer complete interrupt
  uint32_t                :  4;
  uint32_t  LISR_FEIF2    :  1; // Stream 2 FIFO error interrupt
  uint32_t                :  1;
  uint32_t  LISR_DMEIF2   :  1; // Stream 2 direct mode error interrupt
  uint32_t  LISR_TEIF2    :  1; // Stream 2 transfer error interrupt
  uint32_t  LISR_HTIF2    :  1; // Stream 2 half transfer interrupt
  uint32_t  LISR_TCIF2    :  1; // Stream 2 transfer complete interrupt
  uint32_t  LISR_FEIF3    :  1; // Stream 3 FIFO error interrupt
  uint32_t                :  1;
  uint32_t  LISR_DMEIF3   :  1; // Stream 3 direct mode error interrupt
  uint32_t  LISR_TEIF3    :  1; // Stream 3 transfer error interrupt
  uint32_t  LISR_HTIF3    :  1; // Stream 3 half transfer interrupt
  uint32_t  LISR_TCIF3    :  1; // Stream 3 transfer complete interrupt
  uint32_t                :  4;
// 04 High interrupt status register (DMA_HISR)
  uint32_t  HISR_FEIF4    :  1; // Stream 4 FIFO error interrupt
  uint32_t                :  1;
  uint32_t  HISR_DMEIF4   :  1; // Stream 4 direct mode error interrupt
  uint32_t  HISR_TEIF4    :  1; // Stream 4 transfer error interrupt
  uint32_t  HISR_HTIF4    :  1; // Stream 4 half transfer interrupt
  uint32_t  HISR_TCIF4    :  1; // Stream 4 transfer complete interrupt
  uint32_t  HISR_FEIF5    :  1; // Stream 5 FIFO error interrupt
  uint32_t                :  1;
  uint32_t  HISR_DMEIF5   :  1; // Stream 5 direct mode error interrupt
  uint32_t  HISR_TEIF5    :  1; // Stream 5 transfer error interrupt
  uint32_t  HISR_HTIF5    :  1; // Stream 5 half transfer interrupt
  uint32_t  HISR_TCIF5    :  1; // Stream 5 transfer complete interrupt
  uint32_t                :  4;
  uint32_t  HISR_FEIF6    :  1; // Stream 6 FIFO error interrupt
  uint32_t                :  1;
  uint32_t  HISR_DMEIF6   :  1; // Stream 6 direct mode error interrupt
  uint32_t  HISR_TEIF6    :  1; // Stream 6 transfer error interrupt
  uint32_t  HISR_HTIF6    :  1; // Stream 6 half transfer interrupt
  uint32_t  HISR_TCIF6    :  1; // Stream 6 transfer complete interrupt
  uint32_t  HISR_FEIF7    :  1; // Stream 7 FIFO error interrupt
  uint32_t                :  1;
  uint32_t  HISR_DMEIF7   :  1; // Stream 7 direct mode error interrupt
  uint32_t  HISR_TEIF7    :  1; // Stream 7 transfer error interrupt
  uint32_t  HISR_HTIF7    :  1; // Stream 7 half transfer interrupt
  uint32_t  HISR_TCIF7    :  1; // Stream 7 transfer complete interrupt
  uint32_t                :  4;
// 08 Low interrupt flag clear register (DMA_LIFCR)
  uint32_t  LIFCR_CFEIF0  :  1; // Stream 0 clear FIFO error interrupt
  uint32_t                :  1;
  uint32_t  LIFCR_CDMEIF0 :  1; // Stream 0 clear direct mode error interrupt
  uint32_t  LIFCR_CTEIF0  :  1; // Stream 0 clear transfer error interrupt
  uint32_t  LIFCR_CHTIF0  :  1; // Stream 0 clear half transfer interrupt
  uint32_t  LIFCR_CTCIF0  :  1; // Stream 0 clear transfer complete interrupt
  uint32_t  LIFCR_CFEIF1  :  1; // Stream 1 clear FIFO error interrupt
  uint32_t                :  1;
  uint32_t  LIFCR_CDMEIF1 :  1; // Stream 1 clear direct mode error interrupt
  uint32_t  LIFCR_CTEIF1  :  1; // Stream 1 clear transfer error interrupt
  uint32_t  LIFCR_CHTIF1  :  1; // Stream 1 clear half transfer interrupt
  uint32_t  LIFCR_CTCIF1  :  1; // Stream 1 clear transfer complete interrupt
  uint32_t                :  4;
  uint32_t  LIFCR_CFEIF2  :  1; // Stream 2 clear FIFO error interrupt
  uint32_t                :  1;
  uint32_t  LIFCR_CDMEIF2 :  1; // Stream 2 clear direct mode error interrupt
  uint32_t  LIFCR_CTEIF2  :  1; // Stream 2 clear transfer error interrupt
  uint32_t  LIFCR_CHTIF2  :  1; // Stream 2 clear half transfer interrupt
  uint32_t  LIFCR_CTCIF2  :  1; // Stream 2 clear transfer complete interrupt
  uint32_t  LIFCR_CFEIF3  :  1; // Stream 3 clear FIFO error interrupt
  uint32_t                :  1;
  uint32_t  LIFCR_CDMEIF3 :  1; // Stream 3 clear direct mode error interrupt
  uint32_t  LIFCR_CTEIF3  :  1; // Stream 3 clear transfer error interrupt
  uint32_t  LIFCR_CHTIF3  :  1; // Stream 3 clear half transfer interrupt
  uint32_t  LIFCR_CTCIF7  :  1; // Stream 3 clear transfer complete interrupt
  uint32_t                :  4;
// 0C High interrupt flag clear register (DMA_HIFCR)
  uint32_t  HIFCR_CFEIF4  :  1; // Stream 4 clear FIFO error interrupt
  uint32_t                :  1;
  uint32_t  HIFCR_CDMEIF4 :  1; // Stream 4 clear direct mode error interrupt
  uint32_t  HIFCR_CTEIF4  :  1; // Stream 4 clear transfer error interrupt
  uint32_t  HIFCR_CHTIF4  :  1; // Stream 4 clear half transfer interrupt
  uint32_t  HIFCR_CTCIF4  :  1; // Stream 4 clear transfer complete interrupt
  uint32_t  HIFCR_CFEIF5  :  1; // Stream 5 clear FIFO error interrupt
  uint32_t                :  1;
  uint32_t  HIFCR_CDMEIF5 :  1; // Stream 5 clear direct mode error interrupt
  uint32_t  HIFCR_CTEIF5  :  1; // Stream 5 clear transfer error interrupt
  uint32_t  HIFCR_CHTIF5  :  1; // Stream 5 clear half transfer interrupt
  uint32_t  HIFCR_CTCIF5  :  1; // Stream 5 clear transfer complete interrupt
  uint32_t                :  4;
  uint32_t  HIFCR_CFEIF6  :  1; // Stream 6 clear FIFO error interrupt
  uint32_t                :  1;
  uint32_t  HIFCR_CDMEIF6 :  1; // Stream 6 clear direct mode error interrupt
  uint32_t  HIFCR_CTEIF6  :  1; // Stream 6 clear transfer error interrupt
  uint32_t  HIFCR_CHTIF6  :  1; // Stream 6 clear half transfer interrupt
  uint32_t  HIFCR_CTCIF6  :  1; // Stream 6 clear transfer complete interrupt
  uint32_t  HIFCR_CFEIF7  :  1; // Stream 7 clear FIFO error interrupt
  uint32_t                :  1;
  uint32_t  HIFCR_CDMEIF7 :  1; // Stream 7 clear direct mode error interrupt
  uint32_t  HIFCR_CTEIF7  :  1; // Stream 7 clear transfer error interrupt
  uint32_t  HIFCR_CHTIF7  :  1; // Stream 7 clear half transfer interrupt
  uint32_t  HIFCR_CTCIF7  :  1; // Stream 7 clear transfer complete interrupt
  uint32_t                :  4;
// 10 Stream 0 ... 7
  dma_stream_struct S[8];
} dma_struct;

_Static_assert (sizeof(dma_struct) == 208, " dma_struct: bad size");

static dma_struct * const DMA1 = (dma_struct *)0x40026000;
static dma_struct * const DMA2 = (dma_struct *)0x40026400;

