// i2c.h
#pragma once
#include <stdint.h>

typedef volatile struct /* i2c_struct */ {  // word access
// 00 Control register 1 (I2C_CR1)
  uint32_t  CR1_PE         :  1; // Peripheral enable
  uint32_t  CR1_SMBUS      :  1; // SMBus mode
  uint32_t                 :  1; //
  uint32_t  CR1_SMBTYPE    :  1; // SMBus type
  uint32_t  CR1_ENARP      :  1; // ARP enable
  uint32_t  CR1_ENPEC      :  1; // PEC enable
  uint32_t  CR1_ENGC       :  1; // General call enable
  uint32_t  CR1_NOSTRETCH  :  1; // Clock stretching disable (Slave mode)
  uint32_t  CR1_START      :  1; // Start generation
  uint32_t  CR1_STOP       :  1; // Stop generation
  uint32_t  CR1_ACK        :  1; // Acknowledge enable
  uint32_t  CR1_POS        :  1; // Acknowledge/PEC Position (for reception)
  uint32_t  CR1_PEC        :  1; // Packet error checking
  uint32_t  CR1_ALERT      :  1; // SMBus alert
  uint32_t                 :  1; //
  uint32_t  CR1_SWRST      :  1; // Software reset
  uint32_t                 : 16;
// 04 Control register 2 (I2C_CR2)
  uint32_t  CR2_FREQ       :  6; // Peripheral clock frequency
  uint32_t                 :  2;
  uint32_t  CR2_ITERREN    :  1; // Error interrupt enable
  uint32_t  CR2_ITEVTEN    :  1; // Event interrupt enable
  uint32_t  CR2_ITBUFEN    :  1; // Buffer interrupt enable
  uint32_t  CR2_DMAEN      :  1; // DMA requests enable
  uint32_t  CR2_LAST       :  1; // DMA last transfer
  uint32_t                 : 19;
// 08 Own address register 1 (I2C_OAR1)
  uint32_t  OAR1_ADD       : 10; // Interface address, 10-bit mode
                                 //   for 7-bit mode use bits [7:1]
  uint32_t                 :  5;
  uint32_t  OAR1_ADDMODE   :  1; // Addressing mode (slave mode)
  uint32_t                 : 16;
// 0C Own address register 2 (I2C_OAR2)
  uint32_t  OAR2_ENDUAL    :  1; // Dual addressing mode enable
  uint32_t  OAR2_ADD2      :  7; // Interface address
  uint32_t                 : 24;
// 10 Data register (I2C_DR)
  uint32_t  DR;
// 14 Status register 1 (I2C_SR1)
  uint32_t  SR1_SB         :  1; // Start bit (Master mode)
  uint32_t  SR1_ADDR       :  1; // Address sent (master mode)
                                 //   address matched (slave mode)
  uint32_t  SR1_BTF        :  1; // Byte transfer finished
  uint32_t  SR1_ADD10      :  1; // 10-bit header sent (Master mode)
  uint32_t  SR1_STOPF      :  1; // Stop detection (slave mode)
  uint32_t                 :  1;
  uint32_t  SR1_RxNE       :  1; // Data register not empty (receivers)
  uint32_t  SR1_TxE        :  1; // Data register empty (transmitters)
  uint32_t  SR1_BERR       :  1; // Bus error
  uint32_t  SR1_ARLO       :  1; // Arbitration lost (master mode)
  uint32_t  SR1_AF         :  1; // Acknowledge failure
  uint32_t  SR1_OVR        :  1; // Overrun/Underrun
  uint32_t  SR1_PECERR     :  1; // PEC Error in reception
  uint32_t                 :  1;
  uint32_t  SR1_TIMEOUT    :  1; // Timeout or Tlow error
  uint32_t  SR1_SMBALERT   :  1; // SMBus alert
  uint32_t                 : 16;
// 18 Status register 2 (I2C_SR2)
  uint32_t  SR2_MSL        :  1; // Master/slave
  uint32_t  SR2_BUSY       :  1; // Bus busy
  uint32_t  SR2_TRA        :  1; // Transmitter/receiver
  uint32_t                 :  1; //
  uint32_t  SR2_GENCALL    :  1; // General call address (Slave mode)
  uint32_t  SR2_SMBDEFAULT :  1; // SMBus device default address (Slave mode)
  uint32_t  SR2_SMBHOST    :  1; // SMBus host header (Slave mode)
  uint32_t  SR2_DUALF      :  1; // Dual flag (Slave mode)
  uint32_t  SR2_SR2_PEC    :  8; // Packet error checking register
  uint32_t                 : 16;
// 1C Clock control register (I2C_CCR)
  uint32_t  CCR_CCR        : 12; // Clock control in Fm/Sm mode (Master mode)
  uint32_t                 :  2;
  uint32_t  CCR_DUTY       :  1; // Fm mode duty cycle
  uint32_t  CCR_F_S        :  1; // F_S: I2C master mode selection
  uint32_t                 : 16;
// 20 TRISE register (I2C_TRISE)
  uint32_t  TRISE_TRISE    :  6; // Max. rise time in Fm/Sm mode (Master mode)
  uint32_t                 : 26;
} i2c_struct;

_Static_assert (sizeof(i2c_struct) == 36, " i2c_struct: bad size");

static i2c_struct * const I2C1 = (i2c_struct *)0x40005400;
static i2c_struct * const I2C2 = (i2c_struct *)0x40005800;
static i2c_struct * const I2C3 = (i2c_struct *)0x40005C00;

static uint32_t * const I2C1_CR1 = (uint32_t *)0x40005400;
static uint32_t * const I2C1_SR1 = (uint32_t *)0x40005414;
static uint32_t * const I2C1_SR2 = (uint32_t *)0x40005418;

