
ELF    := $(PROG).elf
S19    := $(PROG).s19
HOST   := $(shell hostname)
USER   := $(shell whoami)
TOP    := $(shell dirname `pwd`)
DATE   := $(shell date +%Y%m%d-%H)
TGZ    := $(PROG)-$(DATE).tgz

vpath %.c ..
BLDDIR := ./BUILD
OBJ    := $(patsubst %.c, $(BLDDIR)/%.o, $(SRC))
DEP    := $(patsubst %.c, $(BLDDIR)/%.d, $(SRC))

ifeq ($(HOST),jessi)
	CC   := arm-none-eabi-gcc
	CFV7 := -Wstrict-overflow=1
else  # erst ab gcc 7.0
	CC   := /opt/gcc/install/bin/arm-none-eabi-gcc
	CFV7 := -Wstrict-overflow=3
	CFV7 += -Wimplicit-fallthrough=4 -Wduplicated-branches
endif
OBJCOPY:= arm-none-eabi-objcopy
OBJDUMP:= arm-none-eabi-objdump
SIZE   := arm-none-eabi-size
E2SREC := ../tools/elf-to-srec
UBL    := ../tools/ubl


# ARM Procedure Call Standard for the ARM Architecture
# 7.1.3 Enumerated Types
#   This ABI delegates a choice of representation of enumerated types to a
#   platform ABI (whether defined by a standard or by custom and practice)
#   or to an interface contract if there is no defined platform ABI.
#   The two permitted ABI variants are:
#   * An enumerated type normally occupies a word (int or unsigned int).
#     If a word cannot represent all of its enumerated values the type
#     occupies a double word (long long or unsigned long long).
#   * The type of the storage container for an enumerated type is the
#     smallest integer type that can contain all of its enumerated values.
#   When both the signed and unsigned versions of an integer type can
#   represent all values, this ABI recommends that the unsigned type
#   should be preferred (in line with common practice).
#
#   Discussion
#   The definition of enumerated types in the C and C++ language standards
#   does not define a binary interface and leaves open the following
#   questions.
#   * Does the container for an enumerated type have a fixed size
#     (as expected in most OS environments) or is the size no larger than
#     needed to hold the values of the enumeration (as expected by most
#     embedded users)?
#
# Using the GNU Compiler Collection (For gcc version 7.0.1)
#
# Chapter 4: C Implementation-Defined Behavior
# 4.9 Structures, Unions, Enumerations, and Bit-Fields
#   On some targets, `-fshort-enums' is the default;
#   this is determined by the ABI.
#
# 6.33.1 Common Type Attributes
#   Specifying the `-fshort-enums' flag on the command line is equivalent
#   to specifying the packed attribute on all enum definitions.
#
# 13.8 Certain Changes We Don't Want to Make
#   This would cause storage layout to be incompatible with most other C
#   compilers. And it doesn't seem very important, given that you can get
#   the same result in other ways. The case where it matters most is when
#   the enumeration-valued object is inside a structure, and in that case
#   you can specify a field width explicitly.
#
# https://wiki.debian.org/ArmEabiPort
#
# Enum sizes
#   The EABI defines an optional system for controlling the size of C
#   enumerated types. For arm-linux it was decided to keep the existing
#   behaviour (enums are at least the same size as an int) for consistency
#   with other Linux systems.
#   This is also reflected in the -mabi=aapcs or -mabi=aapcs-linux switches
#   to GCC: aapcs defines enums to be a variable sized type, while with
#   aapcs-linux they are always ints (4 bytes).

CFLAGS := -Wall -Wextra -Wundef -std=c11 -pedantic -g -fno-lto
# -fpic
# -msingle-pic-base -mpic-data-is-text-relative -fomit-frame-pointer
CFLAGS += -Wunused-parameter
CFLAGS += -fno-diagnostics-show-option -fmax-errors=3
CFLAGS += -fstrict-overflow $(CFV7)
CFLAGS += -Wshadow -Wredundant-decls -Wnested-externs
CFLAGS += -Wdisabled-optimization -Wpadded -Wbad-function-cast
# CFLAGS += -Wno-missing-field-initializers
# CFLAGS += -Wconversion
# CFLAGS += -Wstack-usage=72
# CFLAGS += -fstack-usage
CFLAGS += -mthumb -mcpu=cortex-m3 -ffreestanding -falign-functions=4
CFLAGS += -fstrict-volatile-bitfields
# CFLAGS += -fno-short-enums
CFLAGS += -mabi=aapcs
CFLAGS += -Os
# CFLAGS += -save-temps
# CFLAGS += -mtpcs-frame -mtpcs-leaf-frame -fno-omit-frame-pointer
# CFLAGS += -funwind-tables -rdynamic
# CFLAGS += -mpoke-function-name
CFLAGS += -I../include
CFLAGS += -DPROGRAM_NAME="\"$(PROG)\""
CFLAGS += -DHOSTNAME="\"$(HOST)\""
CFLAGS += -DUSERNAME="\"$(USER)\""
CFLAGS += -DBUILD_PATH="\"$(TOP)\""
CFLAGS += -DMINITERM_UART=$(MINITERM_UART)
CFLAGS += $(LOCAL_CFLAGS)

LDMEMORY := ../memory.ld
LDSCRIPT := $(PROG).ld

LDFLAGS := -Wl,--script=$(LDMEMORY)
LDFLAGS += -Wl,--script=$(LDSCRIPT)
LDFLAGS += -nostdlib
LDFLAGS += $(LOCAL_LDFLAGS)
LDFLAGS += -Wl,-Map=$(BLDDIR)/$(PROG).map
# LDFLAGS += -Wl,--gc-sections
# LDFLAGS += -lgcc


$(BLDDIR)/$(S19): $(BLDDIR)/$(ELF)
	$(E2SREC) $(BLDDIR)/$(PROG) $(LDMEMORY) $(LOCAL_VERSION)

$(BLDDIR)/$(ELF): Makefile ../version $(LDMEMORY) $(LDSCRIPT) $(OBJ)
	echo $@
	$(CC) -o $@ $(OBJ) $(LDFLAGS)
	$(SIZE) $@

$(BLDDIR)/%.o: %.c
	echo $<
	$(CC) -c -o $@ $(CFLAGS) $<

$(BLDDIR)/%.d: %.c
	set -e; rm -f $@; \
        $(CC) -MM $(CFLAGS) $< | \
        sed 's,\($*\)\.o[ :]*,$(BLDDIR)/\1.o $@ : ,g' > $@

inter:
	$(UBL) -i

options:
	$(CC) -Q --help=target -c $(CFLAGS)

size:
	$(SIZE) BUILD/$(PROG).elf

run:
	$(UBL) -g BUILD/$(S19)

load:
	$(UBL) BUILD/$(S19)

Dump:
	$(OBJDUMP) -D $(BLDDIR)/$(ELF)

dump:
	$(OBJDUMP) -d -t -C $(BLDDIR)/$(ELF)

srec:
	$(OBJCOPY) -O srec $(BLDDIR)/$(ELF) $(BLDDIR)/$(S19)

clean:
	rm -f $(BLDDIR)/*

-include $(DEP)

.PHONY: tgz RUN run dump srec clean
.SILENT:

