// elf-to-srec.c

#define _POSIX_C_SOURCE 200809L  // strdup
#include <ctype.h>
#include <errno.h>
#include <libgen.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "srecord.h"
#include "fail.h"
#include "../include/image-header.h"

static image_struct dummy;
static const size_t OFF_PRG = offsetof (image_struct, name);
static const size_t SIZ_PRG = sizeof dummy.name;
static const size_t OFF_VER = offsetof (image_struct, version_string);
static const size_t SIZ_VER = sizeof dummy.version_string;
static const size_t OFF_VNO = offsetof (image_struct, verno);
static const size_t OFF_RNO = offsetof (image_struct, revno);
static const size_t OFF_END = offsetof (image_struct, end);
static const size_t OFF_CRC = offsetof (image_struct, crc);
static const size_t OFF_SDE = offsetof (image_struct, source_date_epoch);
static const size_t OFF_CRT = offsetof (image_struct, crt0);
// static const size_t OFF_LIB = offsetof (image_struct, syslib_version);

uint32_t  source_date_epoch (const char *memory_ld);


static void
header_insert_word (img_st *image, size_t offset, uint32_t data)
{
   size_t ads;

   ads = image->hw_ads & image->imask;
   ads += offset;
   image->data[ads + 0] = (data >>  0) & 255;
   image->data[ads + 1] = (data >>  8) & 255;
   image->data[ads + 2] = (data >> 16) & 255;
   image->data[ads + 3] = (data >> 24) & 255;
   return;
}


static void
header_insert_byte (img_st *image, size_t offset, int data)
{
   size_t ads;

   ads = image->hw_ads & image->imask;
   ads += offset;
   image->data[ads] = data & 255;
   return;
}


static void
header_insert_string (img_st *image, size_t offset, int bytes, char *data)
{
   size_t ads;
   int    i1;

   ads = image->hw_ads & image->imask;
   ads += offset;
   for (i1 = 0; i1 < bytes; i1++) {
      if (image->data[ads + i1] != '\0') { return; }
   }
   for (i1 = 0; data[i1] && i1 < bytes - 1; i1++) {
      image->data[ads + i1] = data[i1];
   }
   image->data[ads + i1] = '\0';
   return;
}

#if 0
static void
header_insert_indirect (img_st *image, size_t off1, size_t off2, uint32_t data)
{
   size_t ads1, ads2;

   ads1 = image->hw_ads & image->imask;
   ads1 += off1;

   ads2  = (uint32_t)image->data[ads1 + 0] <<  0;
   ads2 |= (uint32_t)image->data[ads1 + 1] <<  8;
   ads2 |= (uint32_t)image->data[ads1 + 2] << 16;
   ads2 |= (uint32_t)image->data[ads1 + 3] << 24;
   ads2 &= image->imask;
   ads2 += off2;

   image->data[ads2 + 0] = (data >>  0) & 255;
   image->data[ads2 + 1] = (data >>  8) & 255;
   image->data[ads2 + 2] = (data >> 16) & 255;
   image->data[ads2 + 3] = (data >> 24) & 255;
   return;
}
#endif

void
check_header (img_st *image)
{
   if (srec_header (image, OFF_END) != 0) {
      fail (0, "image->end: 0x%X != 0x%X",
            srec_header (image, OFF_END), 0);
   }
   if (srec_header (image, OFF_CRC) != (uint32_t)image->hw_ads) {
      fail (0, "image->crc: 0x%X != 0x%X",
            srec_header (image, OFF_CRC), (uint32_t)image->hw_ads);
   }
   if (srec_header (image, OFF_SDE) != 0) {
      fail (0, "image->sde: 0x%X != 0",
            srec_header (image, OFF_SDE));
   }
}


static uint32_t
crc_stm32 (uint32_t crc, uint32_t data)
{
   int i1;

   crc = crc ^ data;

   for (i1 = 0; i1 < 32; i1++) {
      if (crc & 0x80000000) {
         crc = (crc << 1) ^ 0x04C11DB7; // Polynomial used in STM32
      } else {
         crc = (crc << 1);
      }
   }
   return crc;
}


void
make_crc (img_st *image, int pic_offset)
{
   int      len;
   uint32_t px, crc, data;
   uint32_t crc_ads;

   crc_ads = image->hw_ads + OFF_CRC;

   header_insert_word (image, OFF_CRT,
                       srec_header (image, OFF_CRT) + pic_offset);
   header_insert_word (image, OFF_END, image->end + pic_offset);
   px = image->hw_ads & image->imask;
   crc = 0xFFFFFFFF;

   for (len = image->end - image->hw_ads; len > 0; len -= 4) {
      if (px == (crc_ads & 0x000FFFFF)) {
         px  += 4;
         len -= 4;
      }
      data  = image->data[px++] <<  0;
      data |= image->data[px++] <<  8;
      data |= image->data[px++] << 16;
      data |= image->data[px++] << 24;
      crc = crc_stm32 (crc, data);
   }
   header_insert_word (image, OFF_CRC, crc);
   return;
}


void
elf2srec (char *elfname, char *s19name)
{
   static char  mks19[142];

   fclose (fail_or_open_ro (elfname));  // zwecks einheitlicher Fehlermeldung

   snprintf (mks19, sizeof mks19, "cd %s; arm-none-eabi-objcopy -O srec %s %s",
             dirname (elfname), basename (elfname), basename (s19name));
   system (mks19);
   return;
}


char *
version (char *local_version, int *vno, int *rno)
{
   static const char *valid_chars =
      "0123456789.-ABCDEFGHIJKLMNOPQRSTUVWXYZ-abcdefghijklmnopqrstuvwxyz";
   static char  buf[88];
   FILE        *verh;
   char        *s1, *logl, *version = NULL;

   if (local_version && *local_version) {
      version = local_version;
      logl = "local version";
   } else {
      logl = "global version";
      verh = fail_or_open_ro ("../version");
      while (fgets (buf, sizeof buf, verh)) {
         fail_line += 1;
         if (*buf == '/') { continue; }
         s1 = strchr (buf, '\n');
         if (s1 == NULL) {
            fail (0, "Line too long");
         }
         *s1 = '\0';
         s1 = strchr (buf, '\r');
         if (s1) {
            *s1 = '\0';
         }
         if (strlen (buf) == 0) { continue; }
         if (version != NULL) {
            fail (0, "Multiple version strings '%s', '%s'", version, buf);
         }
         version = strdup (buf);
      }
      fclose (verh);
   }

   if (version == NULL) {
      fail (0, "Version string not found");
   }
   if (strlen (version) >= SIZ_VER) {
      fail (0, "%s too long '%*.*s'", logl, SIZ_VER, SIZ_VER, version);
   }
   for (s1 = version; *s1; s1++) {
      if (strchr (valid_chars, *s1) == NULL) {
         fail (0, "Illegal character '%c' in %s '%s'", *s1, logl, version);
      }
   }
   if (version[0] == 'V' && version[3] == '.'
       && isdigit (version[1]) && isdigit (version[2])
       && isdigit (version[4]) && isdigit (version[5])) {
      *vno = atoi (version + 1);
      *rno = atoi (version + 4);
      if (*vno >= 94 || *rno >= 94) {
         fail (0, "'%s' >= 94 (DAZ limit)", version);
      }
   } else {
      *vno = 0;
      *rno = 0;
   }
   return version;
}


int
main (int argc, char *argv[])
{
   static char  elfname[256], s19name[256], tmpname[256];
   img_st      *img;
   int          vno, rno, pic_offset;

   if (argc != 3 && argc != 4) {
      fail_usage ("program_name memory.ld [local_version]");
   }
   if (getenv ("PIC_OFFSET")) {
      pic_offset = strtol (getenv ("PIC_OFFSET"), NULL, 0);
   } else {
      pic_offset = 0;
   }

   snprintf (elfname, sizeof elfname, "%s.elf", argv[1]);
   snprintf (tmpname, sizeof tmpname, "%s", argv[1]);
   snprintf (s19name, sizeof s19name, "%s.s19", argv[1]);

   elf2srec (elfname, tmpname);
   img = srec_load (tmpname);
   if (unlink (tmpname)) {
      fail (errno, "unlink %s", tmpname);
   }
   check_header (img);

   header_insert_string (img, OFF_PRG, SIZ_PRG, basename (argv[1]));
   header_insert_string (img, OFF_VER, SIZ_VER, version (argv[3], &vno, &rno));
   header_insert_byte (img, OFF_VNO, vno);
   header_insert_byte (img, OFF_RNO, rno);
   header_insert_word (img, OFF_SDE, source_date_epoch (argv[2]));

   make_crc (img, pic_offset);
   srec_print (s19name, img, 16, pic_offset);
   return 0;
}

