// fail.c

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

extern char *program_invocation_short_name;

int                fail_line;

static const char *fail_file;
static dev_t       fail_dev;
static ino_t       fail_ino;


void
fail_usage (const char *fmt, ...)
{
   va_list ap;

   fflush (stdout);
   fprintf (stderr, "Usage: %s ", program_invocation_short_name);

   va_start (ap, fmt);
   vfprintf (stderr, fmt, ap);
   va_end (ap);

   fprintf (stderr, "\n");
   exit (EXIT_FAILURE);
}


void
fail (int errnum, const char *fmt, ...)
{
   va_list ap;

   fflush (stdout);
   fprintf (stderr, "\033[33m" "%s: ", program_invocation_short_name);

   if (errnum != 0) {
      va_start (ap, fmt);
      vfprintf (stderr, fmt, ap);
      va_end (ap);
      fprintf (stderr, " ");
   }

   if (fail_file != NULL) {
      fprintf (stderr, "%s", fail_file);
   }
   if (fail_line > 0) {
      fprintf (stderr, " line %d", fail_line);
   }
   if (fail_file != NULL || fail_line > 0) {
      fprintf (stderr, ": ");
   }

   if (errnum == 0) {
      va_start (ap, fmt);
      vfprintf (stderr, fmt, ap);
      va_end (ap);
   } else {
      fprintf (stderr, "%s", strerror (errnum));
   }
   fprintf (stderr, "\033[m" "\n");
   exit (EXIT_FAILURE);
}


FILE *
fail_or_append (const char *filename)
{
   struct stat  fst;
   FILE        *fp;
   char        *text;

   text = NULL;
   if (stat (filename, &fst)) {
      if (errno != ENOENT) {
         text = "stat";
      }
   } else {
      errno = 0;
      if (! S_ISREG (fst.st_mode)) {
         text = "not a regular file";
      }
      if (fail_dev == fst.st_dev && fail_ino == fst.st_ino) {
         text = "input file is the same as output file";
      }
   }

   if (text == NULL) {
      fp = fopen (filename, "a");
      if (fp == NULL) {
         text = "append";
      }
   }
   if (text != NULL) {
      fail_file = filename;
      fail_line = -1;
      fail (errno, text);
   }
   return fp;
}


FILE *
fail_or_create (const char *filename)
{
   struct stat  fst;
   FILE        *fp;
   char        *text;

   text = NULL;
   if (stat (filename, &fst)) {
      if (errno != ENOENT) {
         text = "stat";
      }
   } else {
      errno = 0;
      if (! S_ISREG (fst.st_mode)) {
         text = "not a regular file";
      }
      if (fail_dev == fst.st_dev && fail_ino == fst.st_ino) {
         text = "input file is the same as output file";
      }
   }

   if (text == NULL) {
      fp = fopen (filename, "w");
      if (fp == NULL) {
         text = "create";
      }
   }
   if (text != NULL) {
      fail_file = filename;
      fail_line = -1;
      fail (errno, text);
   }
   return fp;
}


FILE *
fail_or_open_ro (const char *filename)
{
   struct stat  fst;
   FILE        *fp;

   fail_file = filename;
   fail_line = 0;

   if (stat (filename, &fst)) {
      fail (errno, "stat");
   }
   if (! S_ISREG (fst.st_mode)) {
      fail (0, "not a regular file");
   }
   if (fst.st_size == 0) {
      fail (0, "is empty");
   }
   fail_dev = fst.st_dev;
   fail_ino = fst.st_ino;

   fp = fopen (filename, "r");
   if (fp == NULL) {
      fail_line = -1;
      fail (errno, "open");
   }
   return fp;
}

