// libubl.c

#include <error.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include "fail.h"
#include "libubl.h"
#include "sci.h"

static FILE  *logfile;
static int    newline;


void
ubl_log_open (void)
{
   logfile = fail_or_append ("/tmp/ubl.log");
}


void
ubl_logc (int c1)
{
   static int  c0;

   if (c1 == '\n') {
      c0 = '\0';
      fputc (c1, logfile);
      fflush (logfile);
      newline = 1;
      return;
   }
   newline = 0;
   if (c1 == '\r') {
      c0 = c1;
      return;
   }
   if (c0 == '\r') {
      fputc (c0, logfile);
      c0 = '\0';
   }
   if (c1 == -1) {
      c1 = 255;
   }
   fputc (c1, logfile);
   return;
}


void
ubl_log_printf (const char *format, ...)
{
   static char buf[512];
   va_list     ap;
   char       *s1;

   if (newline == 0) {
      ubl_logc ('\n');
   }
   ubl_logc ('_');
   ubl_logc ('_');
   ubl_logc ('_');
   ubl_logc (' ');
   va_start (ap, format);
   vsnprintf (buf, sizeof buf, format, ap);
   va_end (ap);

   for (s1 = buf; *s1; s1++) {
      ubl_logc (*s1);
   }
   return;
}


static int
ubl_command (int command)
{
   int  ack, err, c1;

   sci_send (command);
   sci_send ((~command) & 255);
   ack = sci_receive (TIME_OUT);
   switch (ack) {
   case -1:
      fprintf (stderr, "ubl_command %s: time-out\n", ubl_cmd_text[command]);
      err = -1;
      break;
   case 0x79:  // ACK
      err = 0;
      break;
   case 0x1F:  // NACK
      fprintf (stderr, "ubl_command %s: NACK\n", ubl_cmd_text[command]);
      err = -1;
      break;
   default:
      err = -1;
      break;
   }
   ack &= 255;

   if (err == 0) {
      ubl_log_printf ("ubl_command %s: %c\n", ubl_cmd_text[command], ack);
   } else {
      ubl_log_printf ("ubl_command %s: 0x%02x ", ubl_cmd_text[command], ack);
      while ((c1 = sci_receive (100)) >= 0) {
         ubl_logc (c1);
      }
      ubl_logc ('\n');
   }
   return err;
}


static int
ubl_address (uint32_t address)
{
   int  checksum, ack, err;

   checksum = ((  address >> 24) & 255)
              ^ ((address >> 16) & 255)
              ^ ((address >>  8) & 255)
              ^ ((address >>  0) & 255);
   sci_send ((address >> 24) & 255);
   sci_send ((address >> 16) & 255);
   sci_send ((address >>  8) & 255);
   sci_send ((address >>  0) & 255);
   sci_send (checksum);
   ack = sci_receive (TIME_OUT);
   switch (ack) {
   case -1:
      fprintf (stderr, "ubl_address 0x%08x: time-out\n", address);
      err = -1;
      break;
   case 0x79:  // ACK
      err = 0;
      break;
   case 0x1F:  // NACK
      fprintf (stderr, "ubl_address 0x%08x: NACK\n", address);
      err = -1;
      break;
   default:
      fprintf (stderr, "ubl_address 0x%08x: garbage (0x%2.2x)\n",
               address, ack & 255);
      err = -1;
      break;
   }
   ubl_log_printf ("ubl_address 0x%08x: 0x%x\n", address, ack & 255);
   return err;
}


static int
ubl_receive (void)
{
   int  data, ack = -1;

   ubl_log_printf ("ubl_receive\n");
   while ((data = sci_receive (TIME_OUT)) != -1) {
      switch (data) {
      case 0x79:
         //    printf (" ACK");
         printf ("%c", data & 0xff);
         ack = 0;
         break;
      case 0x1f:
         printf (" NAK");
         ack = -1;
         break;
      default:
         printf ("%c", data & 0xff);
         ack = -1;
      }
      ubl_logc (data);
   }
   printf ("\n");
   ubl_logc ('\n');
   return ack;
}


int
ubl_connect (int reset)
{
   int  c1;

   printf ("connect\n");
   ubl_log_printf ("ubl_connect\n");
   if (reset) {
      while ((c1 = sci_receive (20)) >= 0) {
         ubl_logc (c1);
      }
      printf ("Break / Reset\n");
      sci_send_break (3000);
      while ((c1 = sci_receive (300)) >= 0) {
         ubl_logc (c1);
      }
      printf ("Boot\n");
   }
   sci_send (0x7f);
   return ubl_receive ();
}


int
ubl_simple_cmd (int cmd)
{
   if (ubl_command (cmd)) { return -1; }
   ubl_receive ();
   return 0;
}


int
ubl_pid (void)
{
   int  data, ack = -1;

   if (ubl_command (UBL_GET_CHIP_ID)) { return -1; }

   data = sci_receive (TIME_OUT);
   ubl_logc (data);
   if (data == -1 || data == 0x1f) { return -1; }

   printf ("Product ID: 0x");

   while ((data = sci_receive (TIME_OUT)) != -1) {
      ubl_logc (data);
      switch (data) {
      case 0x79:
         ack = 0;
         break;
      case 0x1f:
         ack = -1;
         break;
      default:
         printf ("%2.2x", data & 0xff);
         ack = -1;
      }
   }
   printf ("\n");
   return ack;
}


int
ubl_go (uint32_t start)
{
   if (ubl_command (UBL_GO)) { return -1; }
   if (ubl_address (start))  { return -1; }
   ubl_receive ();
   return 0;
}


int
ubl_read_row (uint32_t start, int buf[], int size)
{
   int  col;

   if (size > 256) { size = 256; }
   if (size < 1)               { return -1; }
   if (ubl_command (UBL_READ)) { return -1; }
   if (ubl_address (start))    { return -1; }
   if (ubl_command (size - 1)) { return -1; }

   for (col = 0; col < size; col++) {
      buf[col] = sci_receive (TIME_OUT);
      if (buf[col] == -1) {
         for ( ; col < 16; col++) {
            buf[col] = -1;
         }
         break;
      }
   }
   return 0;
}


int
ubl_read_memory (int s19, uint32_t start, int count)
{
   int  data[16], col, chksum;

   if (count <= 0) {
      count = 256;
   }
   for (/* */; count > 0; start += 16, count -= 16) {
      if (ubl_command (UBL_READ))  { return -1; }
      if (ubl_address (start)) { return -1; }
      if (ubl_command (count > 16 ? 15 : count - 1)) { return -1; }

      for (col = 0; col < 16; col++) {
         data[col] = sci_receive (TIME_OUT);
         if (data[col] == -1) {
            for ( ; col < 16; col++) {
               data[col] = -1;
            }
            break;
         }
      }

      if (s19) {
         chksum = 0x15;
         chksum += (start >> 24) & 255;
         chksum += (start >> 16) & 255;
         chksum += (start >>  8) & 255;
         chksum += (start >>  0) & 255;
         printf ("S315%8.8X", start);
         for (col = 0; col < 16; col++) {
            if (data[col] < 0) {
               printf ("--");
            } else {
               chksum += data[col];
               printf ("%2.2X", data[col]);
            }
         }
         chksum = ~chksum;
         chksum &= 255;
         printf ("%2.2X\r\n", chksum);
      } else {
         printf ("%8.8x ", start);
         for (col = 0; col < 16; col++) {
            if (data[col] < 0) {
               printf (" --");
            } else {
               printf (" %2.2x", data[col]);
            }
            if (col == 7) {
               printf (" ");
            }
         }
         printf ("  |");
         for (col = 0; col < 16; col++) {
            if (data[col] < 32 || data[col] >= 127) {
               data[col] = '.';
            }
            printf ("%c", data[col]);
         }
         printf ("|\n");
      }
   }
   printf ("\n");
   return 0;
}


int
ubl_write_page (uint32_t start, uint8_t *data)
{
   int  checksum;
   int  ack, i1;
   int  count = 256 - 1;

   if (ubl_command (UBL_WRITE))  { return -1; }
   if (ubl_address (start))      { return -1; }
   checksum = count;
   sci_send (count);
   for (i1 = 0; i1 <= count; i1++) {
      checksum ^= *data;
      sci_send (*data++);
   }
   sci_send (checksum);
   ack = sci_receive (TIME_OUT * 2);
   switch (ack) {
   case -1:
      fprintf (stderr, "ubl_write_page 0x%8.8x: time-out\n", start);
      return -1;
   case 0x79:  // ACK
      return 0;
   case 0x1F:  // NACK
      fprintf (stderr, "ubl_write_page 0x%8.8x: NACK\n", start);
      return -1;
   default:
      fprintf (stderr, "ubl_write_page 0x%8.8x: garbage (0x%2.2x)\n",
               start, ack & 255);
      return -1;
   }
   return 0;
}


int
ubl_write_word (uint32_t start, uint32_t data)
{
   int  checksum;
   int  ack;
   int  count = 4 - 1;

   if (ubl_command (UBL_WRITE)) { return -1; }
   if (ubl_address (start))     { return -1; }
   checksum = count
              ^ ((data >> 24) & 255)
              ^ ((data >> 16) & 255)
              ^ ((data >>  8) & 255)
              ^ ((data >>  0) & 255);
   sci_send (count);
   sci_send ((data >>  0) & 255);
   sci_send ((data >>  8) & 255);
   sci_send ((data >> 16) & 255);
   sci_send ((data >> 24) & 255);
   sci_send (checksum);
   ack = sci_receive (TIME_OUT);
   switch (ack) {
   case -1:
      fprintf (stderr, "ubl_write_word 0x%8.8x: time-out\n", start);
      return -1;
   case 0x79:  // ACK
      return 0;
   case 0x1F:  // NACK
      fprintf (stderr, "ubl_write_word 0x%8.8x: NACK\n", start);
      return -1;
   default:
      fprintf (stderr, "ubl_write_word 0x%8.8x: garbage (0x%2.2x)\n",
               start, ack & 255);
      return -1;
   }
   return 0;
}

