// mk-lib-tables.c

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include "fail.h"

static const char *LIB_H = "../include/syslib.h";
static const char *TAB_C = "BUILD/lib-table.c";
static const char *VEC_C = "BUILD/lib-vectors.c";

//extern int  fail_line;

typedef struct {
   char  *symbol;
   int    ads;
} stab_struct;

static stab_struct  svctab[88];
static stab_struct  libtab[88];
static char         bitmap[0x20020000 / 1024];
static const int    FLASH_SIZE = 256;


static char *
string_dup (const char *src)
{
   char  *dst;

   dst = malloc (strlen (src) + 1);
   if (dst == NULL) {
      fail (errno, "string_dup: out of memory");
   }
   return strcpy (dst, src);
}


static int
print_page_char (int mapbit)
{
   switch (mapbit) {
   case  0:
      putchar (' ');
      return 0;
   case  1:
      putchar ('-');
      return 0;
   case  2:
      putchar ('x');
      return 0;
   case  4:
      putchar ('w');
      return 0;
   case 17:
      putchar ('!');
      return 1;
   case 18:
      putchar ('X');
      return 1;
   case 20:
      putchar ('W');
      return 1;
   default:
      putchar (mapbit + '0');
      return 1;
   }
}


static void
check_memory_regions (const char *memory_ld)
{
   static const char  *attr[] = { "(r)", "(rx)", "(rw)", NULL };
   FILE  *map;
   char  *s1, *s2, *s3, *s4, buf[88];
   int    origin, length, page, rwx, errors;

   map = fail_or_open_ro (memory_ld);

   while (fgets (buf, sizeof buf, map)) {
      fail_line += 1;
      if (strstr (buf, "/*")) { continue; }
      s1 = strstr (buf, " ORIGIN = ");
      s2 = strstr (buf, " LENGTH = ");
      if ((s1 != NULL && s2 == NULL) || (s1 == NULL && s2 != NULL)) {
         fail (0, "syntax error");
      }
      if (s1 == NULL && s2 == NULL) { continue; }
      s3 = strchr (buf, '(');
      if (s3) {
         s4 = strchr (s3, ')');
      }
      if (s3 == NULL || s4 == NULL) {
         fail (0, "(rwx) not found");
      }
      *++s4 = '\0';

      for (rwx = 0; attr[rwx]; rwx++) {
         if (strcmp (s3, attr[rwx]) == 0) { break; }
      }
      if (attr[rwx] == NULL) {
         fail (0, "unkown section attribute '%s'", s3);
      }

      origin = strtol (s1 + 9, NULL, 0) / 1024;
      length = strtol (s2 + 9, &s3, 0);
      if (s3 == NULL || *s3 != 'K') {
         fail (0, "unit must be 'K'");
      }
      for (page = origin; page < origin + length; page++) {
         if (page >= (int)sizeof bitmap) {
            fail (0, "page out of range");
         }
         if (bitmap[page] == 0) {
            bitmap[page] |= (1 << rwx);
         } else {
            bitmap[page] |= 16;
         }
      }
   }
   fclose (map);
   fail_line = 0;

   errors = 0;
   for (page = 0; page < 0x08000000 / 1024; page++) {
      if (bitmap[page]) {
         fail (0, "illegal page at %8.8X", page * 1024);
      }
   }
   for ( ; page < 0x08000000 / 1024 + FLASH_SIZE; page++) {
      if (page % 64 == 0) {
         printf ("%8.8X |", page * 1024);
      }
      errors += print_page_char (bitmap[page]);
      if (page && page % 64 == 63) {
         printf ("|\n");
      }
   }
   for (page = 0x08040000 / 1024; page < 0x20000000 / 1024; page++) {
      if (bitmap[page]) {
         fail (0, "illegal page at %8.8X", page * 1024);
      }
   }
   for ( ; page < (int)sizeof bitmap; page++) {
      if (page && page % 64 == 0) {
         printf ("%8.8X |", page * 1024);
      }
      errors += print_page_char (bitmap[page]);
      if (page && page % 64 == 63) {
         printf ("|\n");
      }
   }
   if (errors) {
      fail (0, "overlapping memory regions");
   }
}


static void
print_vectors (char *name, time_t mtime, int svc_width, int sym_width)
{
   char   buf[64];
   FILE  *ld;
   int    i1, blank, svc;

   ld = fail_or_create (VEC_C);
   strftime (buf, sizeof buf, "%F %T", localtime (&mtime));
   fprintf (ld, "// %s %s\n", VEC_C, buf);
   fprintf (ld, "\n");
   fprintf (ld, "// DO NOT EDIT -- generated from %s by %s\n",
            LIB_H, name);
   fprintf (ld, "\n");
   fprintf (ld, "void __attribute__ ((naked, section (\".syslib_table\")))\n"
            "lib (void)\n"
            "{\n");
   fprintf (ld, "   __asm__ (\" SYSLIB_VERSION = 0x%8lX\\n", mtime);
   fprintf (ld, " .global SYSLIB_VERSION\");\n");
   fprintf (ld, "   __asm__ (\"lib: .word 0x%8lX\");\n", mtime);
   fprintf (ld, "   __asm__ (\"     .word 0xFFFFFFFF\");\n");
   fprintf (ld, "\n");
   for (i1 = 0; libtab[i1].symbol; i1++) {
      for (svc = 0; svctab[svc].symbol; svc++) {
         if (strcmp (svctab[svc].symbol, libtab[i1].symbol) == 0) { break; }
      }
      if (svctab[svc].symbol) {
         blank = svc_width - strlen (svctab[svc].symbol);
         fprintf (ld, "   __asm__ (\"");
         fprintf (ld, "%s: %-*.*s", libtab[i1].symbol, blank, blank, " ");
         fprintf (ld, "svc %2d\\n bx lr\\n .word -1\\n", svc);
         fprintf (ld, " .global %-*.*s\");\n",
                  sym_width, sym_width, svctab[svc].symbol);
      } else {
         fprintf (ld, "   __asm__ (\"");
         fprintf (ld, "%-*.*s  ", svc_width, svc_width, " ");
         fprintf (ld, "ldr pc, [pc]\\n             ");
         fprintf (ld, " .word   %-*.*s\");\n",
                  sym_width, sym_width, libtab[i1].symbol);
      }
   }
   fprintf (ld, "}\n");
   fprintf (ld, "\n");
   fclose (ld);
   return;
}


static void
print_symbols (char *progname, time_t mtime, int columns)
{
   char   buf[64];
   FILE  *ld;
   int    i1;

   ld = fail_or_create (TAB_C);
   strftime (buf, sizeof buf, "%F %T", localtime (&mtime));
   fprintf (ld, "// %s %s\n", TAB_C, buf);
   fprintf (ld, "\n");
   fprintf (ld, "// DO NOT EDIT -- generated from %s by %s\n",
            LIB_H, progname);
   fprintf (ld, "\n");
   fprintf (ld, "   __asm__ (\" %-*.*s = 0x%8lX\\n"
            " .global %-*.*s \");\n",
            columns, columns, "SYSLIB_VERSION", mtime,
            columns, columns, "SYSLIB_VERSION");
   fprintf (ld, "\n");
   for (i1 = 0; libtab[i1].symbol; i1++) {
      fprintf (ld, "   __asm__ (\" %-*.*s = 0x%.8x\\n"
               " .global %-*.*s\");\n",
               columns, columns, libtab[i1].symbol,
               libtab[i1].ads, columns, columns, libtab[i1].symbol);
   }
   fprintf (ld, "\n");
   fclose (ld);
   return;
}


int
main (int argc, char *argv[])
{
   struct stat  vst;
   FILE        *src, *map;
   char         buf[256], *symbol;
   char        *s1, *s2;
   int          count, flash_syslib, ads;
   size_t       svc_width, sym_width;

   if (argc != 3 || (strcmp (argv[1], "system") && strcmp (argv[1], "user"))) {
      fail_usage ("system|user memory-xxx.ld");
   }
   check_memory_regions (argv[2]);

   map = fail_or_open_ro (argv[2]);
   flash_syslib = -1;
   while (fgets (buf, sizeof buf, map)) {
      fail_line += 1;
      if (strstr (buf, "flash_syslib")) {
         if ((s1 = strstr (buf, "ORIGIN"))) {
            if ((s1 = strchr (s1, '='))) {
               flash_syslib = strtol (s1 + 1, NULL, 16);
               printf ("flash_syslib: 0x%8X\n", flash_syslib);
            }
         }
      }
   }
   fclose (map);
   fail_line = 0;
   if (flash_syslib == -1) {
      fail (0, "'flash_syslib' not found");
   }


   src = fail_or_open_ro (LIB_H);
   if (stat (LIB_H, &vst)) {
      fail (errno, "stat %s", LIB_H);
   }

   while (fgets (buf, sizeof buf, src)) {
      fail_line += 1;
      if (strcmp (buf,
                  "typedef enum __attribute__ ((packed)) svc_enum {\n") == 0) {
         break;
      }
   }
   if (feof (src)) {
      fail (0, "svc_enum not found");
   }
   svc_width = 0;
   count = 0;
   while (fgets (buf, sizeof buf, src)) {
      fail_line += 1;
      if (strcmp (buf, "} svc_enum;\n") == 0) { break; }
      s1 = strstr (buf, "SVC_");
      s2 = strtok (s1, " ,\n");
      if (s2 == NULL) {
         fail (0, "bad enum definition");
      }
      s2 += 4;
      for (s1 = s2; *s1; s1++) {
         *s1 = tolower (*s1);
      }
      //  printf ("'%s' = %d\n", s2, count);
      if (strlen (s2) > svc_width) {
         svc_width = strlen (s2);
      }
      svctab[count++].symbol = string_dup (s2);
   }
   svctab[count].symbol = NULL;
   if (feof (src)) {
      fail (0, "unexpected EOF");
   }

   ads = flash_syslib + 512;  // size of cortex vector table (fixed)

   sym_width = 0;
   count = 0;
   while (fgets (buf, sizeof buf, src)) {
      fail_line += 1;
      for (s1 = buf; *s1; s1++) {
         switch (*s1) {
         case '\t':
            *s1 = ' ';
            break;
         case '\r':
            *s1 = '\0';
            break;
         case '\n':
            *s1 = '\0';
            break;
         default:
            break;
         }
      }
      for (s1 = buf; *s1 == ' '; s1++) { }
      if (*s1 == '/' || *s1 == '\0') { continue; }
      s2 = strchr (s1, ' ');
      if (s2 == NULL) {
         fail (0, "syntax error");
      }
      symbol = s2;
      while (*symbol == ' ') {
         symbol += 1;
      }
      if (*symbol == '*') {
         symbol += 1;
      }
      s1 = strchr (symbol, ' ');
      if (s1) {
         *s1 = '\0';
      }
      if (strlen (symbol) < 3) {
         fail (0, "syntax error (name too short?)");
      }
      if (strlen (symbol) > sym_width) {
         sym_width = strlen (symbol);
      }
      ads += 8;
      libtab[count].ads = ads;
      libtab[count].symbol = string_dup (symbol);
      count += 1;
   }
   libtab[count].symbol = NULL;
   fclose (src);

   if (strcmp (argv[1], "user") == 0) {
      print_symbols (argv[0], vst.st_mtime, sym_width);
   }
   if (strcmp (argv[1], "system") == 0) {
      print_vectors (argv[0], vst.st_mtime, svc_width, sym_width);
   }
   return 0;
}

