// sci.c

#define _POSIX_C_SOURCE 200809L // strsignal
#include <error.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include "sci.h"

static int  port = -1;
static int  sci_debug;
static int  terminator_signal;


#if 0
static void
terminator (int signal)
{
   terminator_signal = signal;
}
#endif


void
sci_send_break (int milliseconds)
{
   if (milliseconds <   10) {
      milliseconds = 10;
   }
   if (milliseconds > 8000) {
      milliseconds = 8000;
   }
   if (sci_debug) {
      printf ("break %d\n", milliseconds);
   }
   if (tcsendbreak (port, milliseconds) == -1) {
      error (0, errno, "tcsendbreak");
   }
}


int
sci_set_debug (int level)
{
   int  old;

   old = sci_debug;
   sci_debug = level;
   return old;
}


void
sci_send (int c1)
{
   char  zeile[2];

   if (port < 3) { return; }

   zeile[0] = c1;
   write (port, zeile, 1);
   return;
}


int
sci_receive (long milliseconds)
{
   unsigned char  zeile[2];
   struct    timeval  timeout;
   fd_set    rfds;
   int       ready;
   int       rdcnt;
   int       err;

   if (port < 3) { return -1; }

   timeout.tv_sec = milliseconds / 1000;
   timeout.tv_usec = (milliseconds % 1000) * 1000;
   FD_ZERO (&rfds);
   FD_SET (port, &rfds);
   ready = select (port + 1, &rfds, NULL, NULL, &timeout);
   err = errno;
   switch (ready) {
   case 1:
      rdcnt = read (port, zeile, 1);
      if (rdcnt == 1) {
         return zeile[0];
      } else {
         return -1;
      }
   case 0:
      if (sci_debug > 1) {
         printf ("sci_receive: time-out\n");
      }
      errno = 0;
      return -1;
   default:
      if (sci_debug) {
         if (err == EINTR) {
            printf ("sci_receive: select: %s\n", strsignal (terminator_signal));
         } else {
            printf ("sci_receive: select: %s\n", strerror (err));
         }
      }
   }
   errno = err;
   return -1;
}


void
sci_printf (const char *format, ...)
{
   va_list  ap;
   char    *s1, cmd[512];

   va_start (ap, format);
   vsnprintf (cmd, sizeof cmd, format, ap);
   va_end (ap);

   for (s1 = cmd; *s1; s1++) {
      sci_send (*s1);
   }
   return;
}


char *
sci_xfer (int timeout, int rxlines, const char *format, ...)
{
   static char  reply[256], cmd[256];
   va_list      ap;
   char        *s1, c1;
   int          i1;

   va_start (ap, format);
   vsnprintf (cmd, sizeof cmd, format, ap);
   va_end (ap);
   s1 = cmd;

   while (*s1) {
      sci_send (*s1++);
      if ((c1 = sci_receive (timeout)) == -1) {
         return "";
      }
      if (sci_debug) {
         if (c1 != '\r') {
            printf ("%c", c1);
            fflush (stdout);
         }
      }
   }
   sci_send ('\r');
   if (sci_debug) {
      printf ("\n");
      fflush (stdout);
   }

   c1 = sci_receive (timeout);
   if (sci_debug) { printf ("%c", c1); }
   c1 = sci_receive (timeout);
   if (sci_debug) {
      printf ("%c", c1);
      fflush (stdout);
   }

   if (rxlines == -1) { return ""; }

   c1 = sci_receive (timeout);
   if (sci_debug) { printf ("%c", c1); }
   if (c1 == '\r') {
      c1 = sci_receive (timeout);
      if (sci_debug) { printf ("%c", c1); }
   }
   if (sci_debug) { fflush (stdout); }
   if (c1 != '\n') {
      reply[0] = c1;
      i1 = 1;
   } else {
      i1 = 0;
   }
   if (rxlines == 0) { return ""; }

   for (; i1 < 80; i1++) {
      reply[i1] = sci_receive (timeout);
      if (reply[i1] == '\r') {
         reply[i1] = sci_receive (timeout);
      }
      if (reply[i1] == -1) {
         printf ("time-out @3\n");
         return "";
      }
      if (reply[i1] == '\n') {
         reply[i1] = '\0';
         break;
      }
   }
   reply[i1] = '\0';
   if (sci_debug) {
      printf ("sci_xfer reply: '%s'\n", reply);
      fflush (stdout);
   }
   return reply;
}


void
sci_close (void)
{
   if (port != -1) {
      close (port);
      port = -1;
   }
}


int
sci_open (const char *dev, long baud, const char *form)
{
   static int      baudtab[] = {
      0, 50, 75, 110, 134, 150, 200, 300, 600, 1200,
      1800, 2400, 4800, 9600, 19200, 38400, 0
   };
   struct termios  comm_params;
   int             bits;     /* 7 oder 8          */
   int             parity;   /* 'N', 'E' oder 'O' */
   int             stopbits;
   int             strip;
   int             bn;

   for (bn = 1; baudtab[bn]; bn++) {
      if (baudtab[bn] == baud) { break; }
   }
   if (baudtab[bn] == 0) {
      if (baud != 115200) {
         error (0, 0, "sci_open '%s': Baudrate %ld not supported", dev, baud);
      }
      bn = B115200;
   }
   switch (form[0]) {
   case '7':
      bits = CS7;
      strip = ISTRIP;
      break;
   case '8':
      bits = CS8;
      strip = 0;
      break;
   default:
      error (0, 0, "sci_open '%s': '%c' bits not supported", dev, form[0]);
      return -1;
   }
   switch (form[1]) {
   case 'N':
   case 'n':
      parity = 0;
      break;
   case 'E':
   case 'e':
      parity = PARENB;
      break;
   case 'O':
   case 'o':
      parity = PARENB | PARODD;
      break;
   default:
      error (0, 0, "sci_open '%s': parity %c not supported", dev, form[1]);
      return -1;
   }
   switch (form[2]) {
   case '1':
      stopbits = 0;
      break;
   case '2':
      stopbits = CSTOPB;
      break;
   default:
      error (0, 0, "sci_open '%s': %c stopbits not supported", dev, form[2]);
      return -1;
   }
   if ((port = open (dev, O_RDWR | O_NOCTTY | O_NDELAY)) < 1) {
      error (0, errno, "sci_open: open '%s'", dev);
      return -1;
   }
   if (fcntl (port, F_SETFD, FD_CLOEXEC)) {
      error (0, errno, "sci_open: fcntl FD_CLOEXEC '%s'", dev);
      /* return -1; */
   }
   if (ioctl (port, TIOCEXCL, 0)) {
      error (0, errno, "sci_open: ioctl TIOEXCL '%s'", dev);
      close (port);
      port = -1;
      return -1;
   }
   comm_params.c_iflag = IGNBRK | IGNPAR | strip;
   comm_params.c_oflag = 0;
   comm_params.c_cflag = CREAD | CLOCAL | bits | parity | stopbits | bn;
   comm_params.c_lflag = 0;
   comm_params.c_line = 0;
   comm_params.c_cc[VMIN] = 1;
   comm_params.c_cc[VTIME] = 0;
   if (ioctl (port, TCSETS, &comm_params)) {
      error (0, errno, "sci_open: ioctl TCSETS '%s'", dev);
      close (port);
      port = -1;
      return -1;
   }
   return 0;
}

