// ubl.c

#define _POSIX_C_SOURCE 200809L // popen, strdup
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/stat.h>
#include "libubl.h"
#include "sci.h"
#include "srecord.h"

typedef struct {
   char   *cmd;
   char   *p1;
   char   *p2;
   void  (*fun0) (void);
   void  (*fun1) (int);
   void  (*fun2) (int, int);
   char   *text;
} cmd_struct;

extern cmd_struct  cmd_list[];

static img_st *image;


static void
raw_console (int raw)
{
   static struct termios  old_params;
   static struct termios  raw_params;

   switch (raw) {
   case 1:
      raw_params.c_iflag = IGNBRK | IGNPAR;
      raw_params.c_oflag = 0;
      //  raw_params.c_cflag = CREAD | CLOCAL | bits | parity | stopbits | bn;
      raw_params.c_lflag = 0;
      raw_params.c_line = 0;
      raw_params.c_cc[VMIN] = 1;
      raw_params.c_cc[VTIME] = 0;
      if (ioctl (0, TCSETS, &raw_params)) {
         error (0, errno, "console: ioctl TCSETS raw");
         return;
      }
      break;
   case 0:
      if (ioctl (0, TCSETS, &old_params)) {
         error (0, errno, "console: ioctl TCSETS old");
         return;
      }
      break;
   default:
      if (ioctl (0, TCGETS, &old_params)) {
         error (0, errno, "console: ioctl TCGETS old");
         return;
      }
   }
   return;
}


static void
sighandler (int sig __attribute__ ((unused)))
{
   raw_console (0);
   puts ("");
   exit (EXIT_SUCCESS);
}


static char *
find_tty (void)
{
   static char  devpath[64];
   FILE        *pipe;
   char        *s1;

   pipe = popen ("find-ftdi", "r");
   if (pipe == NULL) { return ""; }

   if (fgets (devpath, sizeof devpath, pipe) == NULL) {
      printf ("read from find-ftdi: %s\n", strerror (errno));
      pclose (pipe);
      return "";
   }
   pclose (pipe);
   s1 = strchr (devpath, '\n');
   if (s1) {
      *s1 = '\0';
   }
   return devpath;
}


int
parse_cmd (char *line, cmd_struct command_list[])
{
   char    *s1, *s2, *arg, *cmd = NULL, *eol = NULL;
   int      cx, cn, p1, p2, args, cmdargs;

   p1 = 0;
   p2 = 0;
   for (s1 = line; *s1; s1++) {
      if (*s1 == '\r' || *s1 == '\n' || *s1 == '#') {
         *s1 = '\0';
         eol = s1;
         break;
      }
      if (*s1 == '\t') {
         *s1 = ' ';
      }
      if (cmd == NULL && *s1 != ' ') {
         cmd = s1;
      }
   }
   if (eol == NULL) {
      fprintf (stderr, "Line too long\n");
      return -1;
   }
   if (cmd == NULL) {
      return 0;
   }

   for (s1 -= 1; s1 >= line; s1--) {
      if (*s1 == ' ') {
         *s1 = '\0';
      } else {
         break;
      }
   }

   args = 0;
   arg = strchr (cmd, ' ');
   if (arg) {
      *arg++ = '\0';
      p1 = strtol (arg, &s2, 0);
      args += 1;
      if (s2 && *s2 == ' ') {
         p2 = strtol (s2, &s1, 0);
         s2 = s1;
         args += 1;
      }
      if (s2 && *s2) {
         fprintf (stderr, "Junk at end of line\n");
         return -1;
      }
   }

   for (cx = 0, cn = -1; command_list[cx].cmd; cx++) {
      if (strncmp (cmd, command_list[cx].cmd, strlen (cmd)) == 0) {
         if (cn == -1) {
            cn = cx;
         } else {
            fprintf (stderr, "Ambiguous command: '%s' '%s'\n",
                     command_list[cn].cmd, command_list[cx].cmd);
            return -1;
         }
      }
   }

   if (cn == -1) {
      fprintf (stderr, "Unknown command\n");
      return -1;
   }

   cmdargs = 0;
   if (command_list[cn].p1) {
      args -= 1;
      cmdargs += 1;
   }
   if (command_list[cn].p2) {
      args -= 1;
      cmdargs += 1;
   }
   if (args != 0) {
      fprintf (stderr, "Wrong number of arguments\n");
      return -1;
   }
   switch (cmdargs) {
   case 0:
      command_list[cn].fun0 ();
      break;
   case 1:
      command_list[cn].fun1 (p1);
      break;
   case 2:
      command_list[cn].fun2 (p1, p2);
      break;
   }
   return 0;
}



void
cmd_exit (void)
{
   printf ("\nund tschuess.\n");
   exit (EXIT_SUCCESS);
}


void
cmd_break (int p1)
{
   printf ("%dms break\n", p1);
   sci_send_break (p1);
}


void
cmd_connect (void)
{
   ubl_connect (0);
}


void
cmd_get (void)
{
   ubl_simple_cmd (UBL_GET);
}


void
cmd_version (void)
{
   ubl_simple_cmd (UBL_GET_VERSION);
}


void
cmd_pid (void)
{
   ubl_pid ();
}


void
cmd_read (int p1, int p2)
{
   int  i1;

   if (p2 == 0) {
      for (i1 = 0; i1 < 32 * 4; i1++, p1 += 256) {
         ubl_read_memory (1, p1, p2);
      }
   }
   ubl_read_memory (1, p1, p2);
}


void
cmd_wr_word (int p1, int p2)
{
   ubl_write_word (p1, p2);
}


void
cmd_go (int p1)
{
   static char  last_line[88];
   size_t       lli;
   int          reply, bytes, cmd;

   ubl_log_printf ("cmd_go\n");
// if (p1 == 0) {
//    p1 = 0x20002000;
// }
   ubl_go (p1);
   raw_console (1);
   sci_set_debug (0);
   lli = 0;
   for (;;) {
      reply = sci_receive (100);
      if (reply == -1) {
         if (ioctl (0, FIONREAD, &bytes) == -1) {
            error (EXIT_FAILURE, errno, "ioctl");
         }
         for (/* */; bytes > 0; bytes--) {
            cmd = getchar ();
            if (cmd == ('B' & 31)) {
               raw_console (0);
               return;
            }
            sci_send (cmd);
         }
         if (strncmp (last_line, "../tools/addr-to-line ", 22) == 0) {
            raw_console (0);
            system (last_line);
            return;
         }
      } else {
         fputc (reply, stdout);
         ubl_logc (reply);
         if (reply == '\n') {
            lli = 0;
         } else {
            if (lli < sizeof last_line - 1) {
               last_line[lli++] = reply;
               last_line[lli] = '\0';
            }
         }
      }
   }
}


void
cmd_exec (int p1)
{
   static char  last_line[88];
   size_t       lli;
   int          ads, cmd, reply, bytes, i1, buf[16];

   ubl_log_printf ("cmd_exec\n");
   sci_set_debug (0);
   while (sci_receive (42) != -1) { }
   sci_set_debug (1);

   if (ubl_pid ()) {
      for (i1 = 0; i1 < 2; i1++) {
         if (ubl_connect (1) == 0) { break; }
         sleep (1);
      }
      if (ubl_pid ()) {
         exit (EXIT_FAILURE);
      }
   }

   if (image == NULL) {
      printf ("No image loaded\n");
      return;
   }
   if (image->hw_ads < 0x20002000) {
      if (ubl_read_row (image->hw_ads, buf, 16)) {
         printf ("Read error @ 0x%08x\n", image->hw_ads);
         return;
      }
      for (i1 = 0; i1 < 16; i1++) {
         if (buf[i1] != 255) {
            printf ("\033[33m" "Sector @ 0x%08x not empty\n" "\033[m",
                    image->hw_ads);
            return;
         }
      }
   }
   sci_set_debug (0);
   for (ads = image->hw_ads; ads < image->end; ads += 256) {
      printf ("\r%8.8X ", ads);
      fflush (stdout);
      ubl_write_page (ads, image->data + (ads & image->imask));
   }
   printf ("\r%8.8X \n", ads);

   if (p1 < 1) { return; }

// ubl_go (0x20003000);
   ubl_go (0x08000000);

   if (p1 < 2) { return; }

   raw_console (1);
   sci_set_debug (0);
   lli = 0;
   for (;;) {
      reply = sci_receive (100);
      if (reply == -1) {
         if (ioctl (0, FIONREAD, &bytes) == -1) {
            error (EXIT_FAILURE, errno, "ioctl");
         }
         for (/* */; bytes > 0; bytes--) {
            cmd = getchar ();
            if (cmd == ('B' & 31)) {
               raw_console (0);
               return;
            }
            sci_send (cmd);
         }
         if (strncmp (last_line, "../tools/addr-to-line ", 22) == 0) {
            raw_console (0);
            system (last_line);
            return;
         }
      } else {
         fputc (reply, stdout);
         fflush (stdout);
         ubl_logc (reply);
         if (reply == '\n') {
            lli = 0;
         } else {
            if (lli < sizeof last_line - 1) {
               last_line[lli++] = reply;
               last_line[lli] = '\0';
            }
         }
      }
   }
   return;
}


void
cmd_show (void)
{
// srec_print ("/tmp/ubl.s19");
// gsm_int_print ("/tmp/ubl.gsm_int");
}


void
cmd_fox (void)
{
   sci_printf ("the quick brown fox jumps over the lazy dog\r\n");
   return;
}


void
cmd_help (void)
{
   int  cx;

   for (cx = 0; cmd_list[cx].cmd; cx++) {
      printf ("%s", cmd_list[cx].cmd);
      if (cmd_list[cx].p1) {
         printf ("\t%s", cmd_list[cx].p1);
      }
      if (cmd_list[cx].p2) {
         printf ("\t%s", cmd_list[cx].p2);
      }
      putchar ('\n');
   }
}


cmd_struct  cmd_list[] = {
   { "help",    NULL,  NULL,    cmd_help, 0, 0,    "" },
   { "connect", NULL,  NULL,    cmd_connect, 0, 0, "" },
   { "get",     NULL,  NULL,    cmd_get, 0, 0,     "get available commands" },
   { "version", NULL,  NULL,    cmd_version, 0, 0, "get bootloader version" },
   { "pid",     NULL,  NULL,    cmd_pid, 0, 0,     "get product id" },
   { "read",    "ads", "bytes", 0, 0, cmd_read,    "read memory" },
   { "wword",   "ads", "word",  0, 0, cmd_wr_word, "write memory" },
   { "go",      "ads", NULL,    0, cmd_go, 0,      "start" },
   { "show",    NULL,  NULL,    cmd_show, 0, 0,    "show image buffer" },
   { "break",   "ms",  NULL,    0, cmd_break, 0,   "send break" },
   { "exec",    "ads", NULL,    0, cmd_exec, 0,    "write image & go" },
   { "quit",    NULL,  NULL,    cmd_exit, 0, 0,    "" },
   { NULL,      NULL,  NULL,    NULL, NULL, NULL,  NULL }
//{ "rflash", "offset", "count", &cmd_rflash,  "read flash from 0x08{offset}" },
//{ "rram",   "offset", "count", &cmd_rram,    "read ram from 0x200{offset}" },
};

static struct option  options[] = {
   { "baud",   required_argument, 0, 'b' },
   { "tty",    required_argument, 0, 't' },
   { "run",    no_argument,       0, 'r' },
   { "go",     no_argument,       0, 'g' },
   { "inter",  no_argument,       0, 'i' },
   { NULL,     0,                 0,  0  }
};


int
main (int argc, char *argv[])
{
   static const char       *PIDFILE = "/run/lock/ubl.pid";
   static struct sigaction  sa;
   FILE    *pidfile;
   char    *dl_tty;
   int      dl_baud;
   char     cmdbuf[88], *imagefile;
   int      pid, run, inter, c_opt, ox;

   sa.sa_handler = sighandler;
   sigaction (SIGHUP, &sa, NULL);
   raw_console (42);

   dl_tty = find_tty ();
   dl_baud = 115200;
   imagefile = NULL;
   inter = 0;
   run = 0;

   do {
      c_opt = getopt_long (argc, argv, "b:t:rgi", options, &ox);
      switch (c_opt) {
      case -1:
         if (optind < argc) {
            imagefile = argv[optind];
         }
         break;
      case 'b':
         dl_baud = atoi (optarg);
         break;
      case 't':
         dl_tty = strdup (optarg);
         break;
      case 'g':
         run = 2;
         break;
      case 'r':
         run = 1;
         break;
      case 'i':
         inter = 1;
         break;
      }
   } while (c_opt != -1);

   if (imagefile == NULL && inter == 0) {
      printf ("Usage: ubl [option]... [image.s19]\n"
              "  -b  baud\n"
              "  -t  /dev/ttyUSBx\n"
              "  -g  load image & go\n"
              "  -i  interactive mode\n");
      exit (EXIT_FAILURE);
   }
   printf ("\033[32m%s\033[m  %s %d %d\n", imagefile, dl_tty, dl_baud, run);

   if (sci_open (dl_tty, dl_baud, DL_FORMAT)) {
      pidfile = fopen (PIDFILE, "r");
      if (pidfile) {
         if (fscanf (pidfile, "%d", &pid) == 1) {
            kill (pid, SIGHUP);
            sleep (1);
         }
         fclose (pidfile);
      }
      if (sci_open (dl_tty, dl_baud, DL_FORMAT)) {
         return (EXIT_FAILURE);
      }
   }
   pidfile = fopen (PIDFILE, "w");
   if (pidfile) {
      fprintf (pidfile, "%d\n", getpid ());
      fclose (pidfile);
   }
   ubl_log_open ();

   if (imagefile) {
      ubl_log_printf ("srec_load %s\n", imagefile);
      image = srec_load (imagefile);
      if (image && run >= 0) {
//       if (image->hw_ads < 0x20002000) {
//          printf ("\033[33m" "Not loading %s, hw_ads = %x\n" "\033[m",
//                  imagefile, image->hw_ads);
//          exit (EXIT_FAILURE);
//       }
//       sci_printf ("U\r");
         cmd_exec (run);
         printf ("\n");
      }
   }

   for (; inter > 0; ) {
      printf ("\r* ");
      fgets (cmdbuf, sizeof cmdbuf, stdin);
      if (feof (stdin)) {
         return (EXIT_SUCCESS);
      }
      parse_cmd (cmdbuf, cmd_list);
   }
   return 0;
}

