/*
    ChibiOS/RT - Copyright (C) 2006-2013 Giovanni Di Sirio

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
*/

/*
   Concepts and parts of this file have been contributed by Fabio Utzig,
   chvprintf() added by Brent Roman.
 */

/**
 * @file    chprintf.c
 * @brief   Mini printf-like functionality.
 *
 * @addtogroup chprintf
 * @{
 */

#include <stdarg.h>
#include <stdint.h>
#include <stdbool.h>
#include "syslib.h"

extern volatile task_struct  *current;

static const int MAX_FILLER = 11;

#ifdef CHPRINTF_USE_FLOAT
static char *ltoa_with_divisor (char *p, long num, unsigned radix, long divi)
{
   int i;
   char *q;
   long l, ll;

   l = num;
   if (divi == 0) {
      ll = num;
   } else {
      ll = divi;
   }

   q = p + MAX_FILLER;
   do {
      i = (int)(l % radix);
      i += '0';
      if (i > '9') {
         i += 'A' - '0' - 10;
      }
      *--q = i;
      l /= radix;
   } while ((ll /= radix) != 0);

   i = (int)(p + MAX_FILLER - q);
   do {
      *p++ = *q++;
   } while (--i);

   return p;
}


static char *ltoa (char *p, long num, unsigned radix, int small)
{
   (void) small;
   return ltoa_with_divisor (p, num, radix, 0);
}


static char *ftoa (char *p, double num, int precision)
{
   long l;
   long divisor[] = { 1, 1, 10, 100, 1000, 10000, 100000 };
   double mult[] = { 1.0, 10.0, 100.0, 1000.0, 10000.0, 100000.0, 1000000.0 };

   l = (long)num;
   p = ltoa_with_divisor (p, l, 10, 0);
   *p++ = '.';
   l = (long)((num - l) * mult[precision] + 0.5);
   return ltoa_with_divisor (p, l, 10, divisor[precision]);
}

#else // CHPRINTF_USE_DOUBLE

static char *ltoa (char *p, long num, unsigned radix, int small)
{
   int i;
   char *q;

   q = p + MAX_FILLER;
   do {
      i = (int)(num % radix);
      i += '0';
      if (i > '9') {
         i += ('A' - '0' - 10) | small;
      }
      *--q = i;
      num /= radix;
   } while (num != 0);

   i = (int)(p + MAX_FILLER - q);
   do {
      *p++ = *q++;
   } while (--i);

   return p;
}
#endif // CHPRINTF_USE_FLOAT


/**
 * @brief   System formatted output function.
 * @details This function implements a minimal @p vprintf()-like functionality
 *          with output on a @p BaseSequentialStream.
 *          The general parameters format is: %[-][width|*][.precision|*][l|L]p.
 *          The following parameter types (p) are supported:
 *          - <b>x</b> hexadecimal integer.
 *          - <b>X</b> hexadecimal long.
 *          - <b>o</b> octal integer.
 *          - <b>O</b> octal long.
 *          - <b>d</b> decimal signed integer.
 *          - <b>D</b> decimal signed long.
 *          - <b>u</b> decimal unsigned integer.
 *          - <b>U</b> decimal unsigned long.
 *          - <b>c</b> character.
 *          - <b>s</b> string.
 *          .
 *
 * @param[in] chp       pointer to a @p BaseSequentialStream implementing object
 * @param[in] fmt       formatting string
 * @param[in] ap        list of parameters
 *
 * @api
 */
int
vprintf (const char *fmt, va_list ap)
{
   char *p, *s, c, filler;
   int i, precision, width, count = 0;
   bool is_long, left_align;
   int  small;
   long l;
#ifdef CHPRINTF_USE_FLOAT
   double f;
   char tmpbuf[2 * MAX_FILLER + 1];
#else
   char tmpbuf[MAX_FILLER + 1];
#endif
// const char *tmp;
// int lang;

// lang = 0;
// for (tmp = fmt; *tmp; tmp++) {
//    if (*tmp == '^') {
//       if (++lang == current->lc_lang) {
//          fmt = tmp + 1;
//          break;
//       }
//    }
// }

   while (true) {
      c = *fmt++;
      if (c == 0 || c == '^') {
         return count;
      }
      if (c != '%') {
         putchar ((uint8_t)c);
         count++;
         continue;
      }
      p = tmpbuf;
      s = tmpbuf;
      left_align = false;
      small = 0;
      if (*fmt == '-') {
         fmt++;
         left_align = true;
      }
      filler = ' ';
      if ((*fmt == '.') || (*fmt == '0')) {
         fmt++;
         filler = '0';
      }
      width = 0;
      while (true) {
         c = *fmt++;
         if (c >= '0' && c <= '9') {
            c -= '0';
         } else {
            if (c == '*') {
               c = va_arg (ap, int);
            } else {
               break;
            }
         }
         width = width * 10 + c;
      }
      precision = 0;
      if (c == '.') {
         while (true) {
            c = *fmt++;
            if (c >= '0' && c <= '9') {
               c -= '0';
            } else {
               if (c == '*') {
                  c = va_arg (ap, int);
               } else {
                  break;
               }
            }
            precision *= 10;
            precision += c;
         }
      }
      /* Long modifier.*/
      if (c == 'l' || c == 'L') {
         is_long = true;
         if (*fmt) {
            c = *fmt++;
         }
      } else {
         is_long = false;
      }

      /* Command decoding.*/
      switch (c) {
      case 'c':
         filler = ' ';
         *p++ = va_arg (ap, int);
         break;
      case 's':
         filler = ' ';
         if ((s = va_arg (ap, char *)) == 0) {
            s = "(null)";
         }
//       lang = 0;
//       for (tmp = s; *tmp; tmp++) {
//          if (*tmp == '^') {
//             if (++lang == current->lc_lang) {
//                s = (char *)(tmp + 1);
//                break;
//             }
//          }
//       }
         if (precision == 0) {
            precision = 32767;
         }
         for (p = s; *p && (--precision >= 0); p++) {
            ;
         }
         break;
      case 'd':
      case 'i':
         if (is_long) {
            l = va_arg (ap, long);
         } else {
            l = va_arg (ap, int);
         }
         if (l < 0) {
            *p++ = '-';
            l = -l;
         }
         p = ltoa (p, l, 10, 0);
         break;
#ifdef CHPRINTF_USE_FLOAT
      case 'f':
         f = va_arg (ap, double);
         if (f < 0) {
            *p++ = '-';
            f = -f;
         }
         p = ftoa (p, f, precision);
         break;
#endif
      case 'p':
         putchar ('0');
         putchar ('x');
         count += 2;
         is_long = true;
      /* FALLTHRU */
      case 'x':
         small = 'a' - 'A';
      /* FALLTHRU */
      case 'X':
         c = 16;
         goto unsigned_common;
      case 'u':
         c = 10;
         goto unsigned_common;
      case 'o':
         c = 8;
unsigned_common:
         if (is_long) {
            l = va_arg (ap, unsigned long);
         } else {
            l = va_arg (ap, unsigned int);
         }
         p = ltoa (p, l, c, small);
         break;
      default:
         *p++ = c;
         break;
      }
      i = (int)(p - s);
      if ((width -= i) < 0) {
         width = 0;
      }
      if (left_align == false) {
         width = -width;
      }
      if (width < 0) {
         if (*s == '-' && filler == '0') {
            putchar ((uint8_t)*s++);
            count++;
            i--;
         }
         do {
            putchar ((uint8_t)filler);
            count++;
         } while (++width != 0);
      }
      while (--i >= 0 && *s != '^') {
         putchar ((uint8_t)*s++);
         count++;
      }
      while (width) {
         putchar ((uint8_t)filler);
         count++;
         width--;
      }
   }
}


int
printf (const char *fmt, ...)
{
   va_list ap;
   int     count;

   va_start (ap, fmt);
   count = vprintf (fmt, ap);
   va_end (ap);
   return count;
}

/** @} */

